import re

import astunparse
import ast
import attr

from .renderer import render_with_identifiers, parse_identifiers
from .skeleton import Skeleton, Blank, Given


@attr.s
class Reskeletonizer:
    ignore_whitespace = attr.ib(default=False)
    normalizer = attr.ib(default=None)
    """
    Reskeletonizer that uses character-level matching to match given portions the skeleton
        in the provided code, in order to find blanks.

    Arguments
        ignore_whitespace: whether to ignore whitespace when resolving blanks. Default False
        normalizer: function that parses and unparses code, to remove extraneous formatting.
            Default is None, or no normalization. If this is used, the returned reskeletonized
            code will correspond to the normalized code rather than the original code.
            
            For this to work, the blanks in the original skeleton must correspond to places an
            identifier could be placed, syntactically. Additionally, any sequence of [a-z]*
            must be a valid identifier for this to work.
    """

    def reskeletonize(self, skeleton: Skeleton, code: str) -> Skeleton:
        if self.normalizer is not None:
            code = self.normalizer(code)
            skeleton_code, skeleton_ids = render_with_identifiers(skeleton)
            skeleton_code = self.normalizer(skeleton_code)
            skeleton = parse_identifiers(skeleton_code, skeleton_ids)

        match = self.create_regex(skeleton).match(code)
        if not match:
            raise CannotReskeletonizeException

        blanks = match.groups()
        assert len(blanks) == len(skeleton.segments)

        new_segments = []
        for segment, content in zip(skeleton.segments, blanks):
            if isinstance(segment, Blank):
                new_segments.append(Blank(content))
            elif isinstance(segment, Given):
                new_segments.append(Given(content))
            else:
                raise AssertionError("Should be unreachable")

        return Skeleton(new_segments)

    def create_regex(self, skeleton):
        regex_chunks = ["^"]
        for segment in skeleton.segments:
            regex_chunks.append(segment.matcher_regex(self._match_given_text))
        regex_chunks.append("$")
        pattern = "".join(regex_chunks)
        return re.compile(pattern, re.DOTALL)

    def _match_given_text(self, code):
        if not self.ignore_whitespace:
            return re.escape(code)
        return r"\s+".join(re.escape(word) for word in re.split(r"\s", code))


def normalize_python(code):
    return astunparse.unparse(ast.parse(code, "<<code>>"))


class CannotReskeletonizeException(Exception):
    pass



from abc import ABC, abstractmethod

import attr

from .renderer import SkeletonRenderer


@attr.s
class Skeleton:
    segments = attr.ib()

    def render(self, renderer: SkeletonRenderer):
        return renderer.combine([segment.render(renderer) for segment in self.segments])


class Segment(ABC):
    @abstractmethod
    def render(self, renderer: SkeletonRenderer):
        pass

    @abstractmethod
    def matcher_regex(self, quotation):
        pass


@attr.s
class Blank(Segment):
    solution = attr.ib()

    def render(self, renderer: SkeletonRenderer):
        return renderer.render_blank(self)

    def matcher_regex(self, quotation):
        return r"(.*?)"


@attr.s
class Given(Segment):
    code = attr.ib()

    def render(self, renderer: SkeletonRenderer):
        return renderer.render_given(self)

    def matcher_regex(self, quotation):
        return "(" + quotation(self.code) + ")"



import random
import re
from abc import ABC, abstractmethod

import attr


class SkeletonRenderer(ABC):
    @abstractmethod
    def combine(self, per_segment_outputs):
        pass

    @abstractmethod
    def render_blank(self, blank):
        pass

    @abstractmethod
    def render_given(self, given):
        pass


@attr.s
class UnderscoreBlankRenderer(SkeletonRenderer):
    blank_size = attr.ib(default=6)

    def combine(self, per_segment_outputs):
        return "".join(per_segment_outputs)

    def render_blank(self, blank):
        return "_" * self.blank_size

    def render_given(self, given):
        return given.code


@attr.s
class DisplaySolutionsRenderer(SkeletonRenderer):
    start = attr.ib(default="<<<")
    end = attr.ib(default=">>>")

    def combine(self, per_segment_outputs):
        return "".join(per_segment_outputs)

    def render_blank(self, blank):
        return self.start + blank.solution + self.end

    def render_given(self, given):
        return given.code


class IdentifierBlankRenderer(SkeletonRenderer):
    """
    Renders each blank as an identifier, as determined by the provided blanks_map.

    Arguments:
        blanks_map: mapping from id(blank) -> identifier
    """

    def __init__(self, blanks_map):
        self.blanks_map = blanks_map

    def combine(self, per_segment_outputs):
        return "".join(per_segment_outputs)

    def render_blank(self, blank):
        return self.blanks_map[id(blank)]

    def render_given(self, given):
        return given.code


def render_with_identifiers(skeleton, identifier_length=30):
    """
    Render the given skeleton with random identifiers per blank. The identifiers are each
    drawn from the set {a...z}^identifier_length

    :param skeleton: a Skeleton object containing givens and blanks.
    :param identifier_length: the length of the identifiers to insert
    :return: the code with the inserted identifiers, and a mapping from each identifier to its corresponding blank
    """
    from skeletonize.skeleton import Blank

    def generate_id():
        return "".join(
            chr(ord("a") + random.randint(0, 25)) for _ in range(identifier_length)
        )

    identifiers = {
        generate_id(): blank for blank in skeleton.segments if isinstance(blank, Blank)
    }

    rendered_code = skeleton.render(
        IdentifierBlankRenderer(
            {id(blank): identifier for identifier, blank in identifiers.items()}
        )
    )

    return rendered_code, identifiers


class DuplicateIdentifierException(Exception):
    pass


def parse_identifiers(code, identifier_to_blank):
    """
    Parses the given code into a skeleton, using each identifier in identifier_to_blank and converting it to the
    corresponding blank

    :param code: the code to process and convert to a skeleton
    :param identifier_to_blank: a mapping from identifier -> Blank object
    :return: a Skeleton containing Given sections from the code and Blank sections from identifiers_for_blank
    """
    from .parser import SkeletonParser
    from .skeleton import Skeleton, Given

    if not identifier_to_blank:
        return Skeleton([Given(code)])

    pattern = re.compile(
        "|".join(re.escape(identifier) for identifier in identifier_to_blank)
    )
    used = set()

    def substitution(match):
        identifier = match.group()
        if identifier in used:
            raise DuplicateIdentifierException(identifier)
        used.add(identifier)
        solution = identifier_to_blank[identifier].solution
        return "\x00" * 3 + solution + "\x01" * 3

    code = pattern.sub(substitution, code)
    return SkeletonParser(start_char="\x00", end_char="\x01").parse(code)



import re
from cached_property import cached_property

from skeletonize.skeleton import Given, Skeleton, Blank


class SkeletonParser:
    def __init__(self, start_char="<", end_char=">", number_to_match=3):
        self.start_char = start_char
        self.end_char = end_char
        self.number_to_match = number_to_match

    def parse(self, skeleton_code):
        self._check_too_long(skeleton_code)

        chunks = []

        def add_given(text):
            self._check_no_delimiters(text)
            if text:
                chunks.append(Given(text))

        def add_blank(text):
            self._check_no_delimiters(text)
            chunks.append(Blank(text))

        index = 0
        for match in self._blank_pattern.finditer(skeleton_code):
            add_given(skeleton_code[index : match.start()])
            add_blank(match.group(1))
            index = match.end()

        add_given(skeleton_code[index:])
        return Skeleton(chunks)

    @cached_property
    def _blank_pattern(self):
        start = re.escape(self.start_char * self.number_to_match)
        end = re.escape(self.end_char * self.number_to_match)
        return re.compile(r"{}(.*?){}".format(start, end), flags=re.M | re.S)

    def _bad_pattern(self, count):
        start = re.escape(self.start_char * count)
        end = re.escape(self.end_char * count)
        return re.compile(r"{}|{}".format(start, end), flags=re.M | re.S)

    def _check_too_long(self, skeleton_code):
        match = self._bad_pattern(self.number_to_match + 1).findall(skeleton_code)
        if match:
            raise TooLongDelimiterException(str(match))

    def _check_no_delimiters(self, text):
        match = self._bad_pattern(self.number_to_match).findall(text)
        if match:
            raise ExtraDelimiterException(str(match))


class TooLongDelimiterException(Exception):
    pass


class ExtraDelimiterException(Exception):
    pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="skeletonize",
    version="0.1",
    author="Kavi Gupta",
    author_email="skeletonize@kavigupta.org",
    description="Handles skeletonization and deskeletonization of python code.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kavigupta/skeletonize",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=[]
)
